﻿using System;
using System.IO;
using System.Text;
using PacketDotNet;


namespace LogicGuard.Public
{
  /// <summary>
  /// Message implementation.
  /// </summary>
  /// <typeparam name="T"></typeparam>
  public class Message
  {
    public Message(object pl, long ts)
    {
      this.Payload = pl;
      this.Timestamp = ts;
    }

    public override string ToString()
    {
      string msg = "Null";
      if (Payload != null) msg = Payload.ToString();
      return msg + '#' + Timestamp.ToString();
    }

    public object Payload { get; set; }
    public long Timestamp { get; set; }
  }

  public class PrimitiveValue<T>
  {
    private T value;
    private long timestamp = 0;

    public PrimitiveValue(T value, long timestamp)
    {
      this.value = value;
      this.timestamp = timestamp;
    }

    public T Value
    {
      get { return value; }
    }

    public long Timestamp
    {
      get { return timestamp; }
    }

    public override string ToString()
    {
      return "Value: " + value + "\tTimestamp: " + timestamp;
    }
  }

  // ---------------------------------------------------------------------------
  // TCP/IP message.
  // it will be the Payload entity of Message. 
  // ---------------------------------------------------------------------------
  public class TcpIpPacket
  {
    // the addresses
    public string SourceAddress { get; set; }
    public ushort SourcePort { get; set; }
    public string DestinationAddress { get; set; }
    public ushort DestinationPort { get; set; }

    // the content
    public byte[] Payload { get; set; }

    // Pcap packet
    public Packet Packet { get; set; }
    public DateTime TimeStamp { get; set; }

    // some flags
    public bool Ack { get; set; }
    public bool Fin { get; set; }
    public bool Syn { get; set; }
    public bool Rst { get; set; }

    public override string ToString()
    {
      return "from " + this.SourceAddress + ":" + this.SourcePort +
    " to " + this.DestinationAddress + ":" + this.DestinationPort +
    " (" + this.Payload.Length + " bytes)";
    }
  }

  public class PcapPacket
  {
    public DateTime TimeStamp { get; private set; }
    public Packet Packet { get; private set; }

    public PcapPacket(Packet packet, DateTime ts)
    {
      TimeStamp = ts;
      Packet = packet;
    }

    public override string ToString()
    {
      return "@" + TimeStamp + ":" + Packet;
    }
  }

  // ---------------------------------------------------------------------------
  // ICMP message.
  // it will be the Payload entity of Message. 
  // ---------------------------------------------------------------------------
  public class IcmpPacket
  {
    // the addresses
    public string SourceAddress { get; set; }
    public string DestinationAddress { get; set; }

    // types
    public IpVersion Version { get; set; }
    public string Type { get; set; }

    // Pcap packet
    public Packet Packet { get; set; }
    public DateTime TimeStamp { get; set; }

    public bool IsEchoRequst { get; set; }
    public bool IsEchoRepy { get; set; }
    public bool IsEcho { get; set; }

    public override string ToString()
    {
      return "ICMP -> Type: " + Type + " Version: " + Version + " from " + SourceAddress +
    " to " + DestinationAddress + " @ " + TimeStamp;
    }
  }

  /// <summary>
  /// The DNS TYPE (RFC1035 3.2.2/3) - 4 types are currently supported. Also, I know that this
  /// enumeration goes against naming guidelines, but I have done this as an ANAME is most
  /// definetely an 'ANAME' and not an 'Aname'
  /// </summary>
  public enum DnsType
  {
    None = 0, ANAME = 1, NS = 2, SOA = 6, MX = 15
  }

  /// <summary>
  /// The DNS CLASS (RFC1035 3.2.4/5)
  /// Internet will be the one we'll be using (IN), the others are for completeness
  /// </summary>
  public enum DnsClass
  {
    None = 0, IN = 1, CS = 2, CH = 3, HS = 4
  }

  /// <summary>
  /// (RFC1035 4.1.1) These are the return codes the server can send back
  /// </summary>
  public enum ReturnCode
  {
    Success = 0,
    FormatError = 1,
    ServerFailure = 2,
    NameError = 3,
    NotImplemented = 4,
    Refused = 5,
    Other = 6
  }

  /// <summary>
  /// (RFC1035 4.1.1) These are the Query Types which apply to all questions in a request
  /// </summary>
  public enum Opcode
  {
    StandardQuery = 0,
    InverseQuerty = 1,
    StatusRequest = 2,
    Reserverd3 = 3,
    Reserverd4 = 4,
    DynamicUpdate = 5,
    Reserverd6 = 6,
    Reserverd7 = 7,
    Reserverd8 = 8,
    Reserverd9 = 9,
    Reserverd10 = 10,
    Reserverd11 = 11,
    Reserverd12 = 12,
    Reserverd13 = 13,
    Reserverd14 = 14,
    Reserverd15 = 15,
  }

  /// <summary>
  /// DNS Packet
  /// </summary>
  public class DnsPacket
  {
    //More information 
    // https://technet.microsoft.com/en-us/library/dd197470(v=ws.10).aspx
    private readonly byte[] data;
    private readonly string src;
    private readonly string dst;
    private readonly short tid;
    private readonly short length;
    private readonly bool resp;
    private readonly bool reqs;
    private readonly bool authoritativeAnswer;
    private readonly bool recursionAvailable;
    private readonly bool truncated;

    private readonly ReturnCode returnCode;
    private readonly Opcode opCode;
    private readonly DateTime dt;

    public bool IsResponse { get { return resp; } }
    public bool IsRequest { get { return reqs; } }
    public bool IsTruncated { get { return truncated; } }
    public bool IsAuthoritativeAnswer { get { return authoritativeAnswer; } }
    public bool IsRecursionAvailable { get { return recursionAvailable; } }
    public short TransID { get { return tid; } }
    public short Length { get { return length; } }
    public string SourceIP { get { return src; } }
    public string DestinationIP { get { return dst; } }
    public ReturnCode ReturnCode { get { return returnCode; } }
    public Opcode OpCode { get { return opCode; } }
    public DateTime Timestamp { get { return dt; } }

    public DnsPacket(string src, string dst, byte[] data, DateTime dt)
    {
      this.src = src;
      this.dst = dst;
      this.data = data;
      this.dt = dt;

      //parsing
      length = GetShort(data, 0);
      tid = GetShort(data, 2);

      byte flags1 = data[4];
      byte flags2 = data[5];

      resp = GetBit(flags1, 1);
      reqs = !resp;

      // get return code from lowest 4 bits of byte 3
      int rc = flags2 & 15;
      // if its in the reserved section, set to other
      if (rc > 6) rc = 6;
      returnCode = (ReturnCode)rc;

      // other bit flags
      authoritativeAnswer = ((flags1 & 4) != 0);
      recursionAvailable = ((flags2 & 128) != 0);
      truncated = ((flags1 & 2) != 0);

      //Opcode
      int oc = flags2 & 1;
      oc = oc << 3;
      oc = ((flags1 >> 3) & 7);
      opCode = (Opcode)oc;
    }

    public override string ToString()
    {
      var sb = new StringBuilder();
      sb.Append("Src: " + SourceIP);
      sb.Append("\tDst: " + DestinationIP);
      sb.Append(String.Format("\nTransationID: 0x{0:X2}", TransID));
      sb.Append(String.Format("\nLength: " + length));
      if (IsResponse)
        sb.Append("\nType: Response");
      else
        sb.Append("\nType: Request");

      sb.Append("\nAuthoritative Answer: " + authoritativeAnswer);
      sb.Append("\nRrecursion Available : " + recursionAvailable);
      sb.Append("\nTruncated: " + truncated);
      sb.Append("\nOpCode: " + opCode);
      sb.Append("\nReturn Code: " + returnCode);
      /*
      sb.Append("\nBytes: ");
      foreach (byte b in data)
        sb.Append(String.Format("{0:X2} ", b));
      */
      return sb.ToString();
    }

    private static bool GetBit(byte b, int bitNumber)
    {
      return (b & (1 << bitNumber - 1)) != 0;
    }

    private static short GetShort(byte[] message, int position)
    {
      return (short)(message[position] << 8 | message[position + 1]);
    }
  }

  /// <summary>
  /// DHCP packets types.
  /// </summary>
  public enum DhcpType : byte
  {
    DhcpDiscover = 1,
    DhcpOffer = 2,
    DhcpRequest = 3,
    DhcpDecline = 4,
    DhcpACK = 5,
    DhcpNAK = 6,
    DhcpRelease = 7,
    DhcpInform = 8
  }

  /// <summary>
  /// DHCP Packet
  /// </summary>
  public class DhcpPacket
  {
    private readonly byte[] _data;
    private string serverIp = "?";
    private string leaseTime = "?";
    private string renewalTime = "?";
    private string rebindingTime = "?";
    private string subnetMask = "?";
    private string routers = "?";
    private string dnsServers = "?";
    private string hostName = "?";
    private string domainName = "?";
    private byte otype = 0;
    private uint xid;
    private string loggingText = "";

    private readonly DateTime _dt;

    public DhcpPacket(byte[] data, DateTime dt)
    {
      _data = data;
      _dt = dt;
      parse();
    }

    public DateTime Timestamp
    {
      get { return _dt; }
    }

    public string ServerIp
    {
      get { return serverIp; }
    }

    public string LeaseTime
    {
      get { return leaseTime; }
    }

    public string RenewalTime
    {
      get { return renewalTime; }
    }

    public string RebindingTime
    {
      get { return rebindingTime; }
    }

    public string SubnetMask
    {
      get { return subnetMask; }
    }

    public string Routers
    {
      get { return routers; }
    }

    public string DnsServers
    {
      get { return dnsServers; }
    }

    public string HostName
    {
      get { return hostName; }
    }

    public string DomainName
    {
      get { return domainName; }
    }

    public byte Otype
    {
      get { return otype; }
    }

    // xid chosen by client
    public uint Tid
    {
      get { return xid; }
    }

    public byte[] Data
    {
      get { return _data; }
    }

    private void parse()
    {
      //Field      DHCPOFFER            
      //-----      ---------            
      //'op'       BOOTREPLY            
      //'htype'    (From "Assigned Numbers" RFC)
      //'hlen'     (Hardware address length in octets)
      //'hops'     0                    
      //'xid'      'xid' from client DHCPDISCOVER message              
      //'secs'     0
      //'ciaddr'   0
      //                                
      //'yiaddr'   IP address offered to client            
      //'siaddr'   IP address of next bootstrap server     
      //'flags'    'flags' from client DHCPDISCOVER message
      //'giaddr'   'giaddr' from client DHCPDISCOVER message              
      //'chaddr'   'chaddr' from client DHCPDISCOVER message              
      //'sname'    Server host name or options
      //'file'     Client boot file name or options
      //'options'  options

      var binaryReader = new BinaryReader(new MemoryStream(_data));

      // skip to xid
      binaryReader.ReadBytes(4);

      // xid chosen by client
      xid = binaryReader.ReadUInt32();

      // skip to ciaddr
      binaryReader.ReadBytes(4);
      string ciaddr = ReadIpv4(binaryReader);
      string yiaddr = ReadIpv4(binaryReader);

      // skip 220 bytes to options
      binaryReader.ReadBytes(220);

      while (binaryReader.BaseStream.Position < binaryReader.BaseStream.Length)
      {
        byte optionCode = binaryReader.ReadByte();

        if (optionCode != 0 && optionCode != 255)
        {
          byte optionLength = binaryReader.ReadByte();

          switch (optionCode)
          {
            // Subnet Mask
            case 1:
              subnetMask = ReadIpv4(binaryReader);
              break;

            case 3:
              routers = ReadIpv4List(optionLength, binaryReader);
              break;

            case 6:
              dnsServers = ReadIpv4List(optionLength, binaryReader);
              break;

            case 12:
              if (optionLength > 0)
              {
                byte[] buffer = binaryReader.ReadBytes(optionLength);
                if (buffer.Length != optionLength)
                  Console.WriteLine("WARNING: Could not read all bytes, option 12");
                hostName = Encoding.ASCII.GetString(buffer, 0, buffer.Length);
              }
              break;

            case 15:
              if (optionLength > 0)
              {
                byte[] buffer = binaryReader.ReadBytes(optionLength);
                if (buffer.Length != optionLength)
                  Console.WriteLine("WARNING: Could not read all bytes, option 15");
                domainName = Encoding.ASCII.GetString(buffer, 0, buffer.Length - 1);
              }
              break;

            // lease time
            case 51:
              leaseTime = ReadTimeInterval(binaryReader);
              break;

            // DHCP message type
            case 53:
              otype = binaryReader.ReadByte();
              //1     DHCPDISCOVER
              //2     DHCPOFFER
              //3     DHCPREQUEST
              //4     DHCPDECLINE
              //5     DHCPACK
              //6     DHCPNAK
              //7     DHCPRELEASE
              //8     DHCPINFORM
              break;

            // DHCP server identifier
            case 54:
              serverIp = ReadIpv4(binaryReader);
              break;

            case 58:
              renewalTime = ReadTimeInterval(binaryReader);
              break;

            case 59:
              rebindingTime = ReadTimeInterval(binaryReader);
              break;

            case 24: // MTU aging, ignore
            case 46: // Netbios stuff ignore
            case 81: // Client FQDN feature ignored
            case 254: // extensions, ignore
              binaryReader.ReadBytes(optionLength);
              break;

            default:
              {
                loggingText += string.Format("\nOption {0} not supported yet (length={1})", optionCode, optionLength);
                binaryReader.ReadBytes(optionLength);
                break;
              }
          }
        }
      }

      if (loggingText != "") loggingText += "\n";

      switch (otype)
      {
        case 1:
          loggingText += string.Format("DISCOVER\t{0:X8}\t{1}", xid, ciaddr);
          break;
        case 2:
          loggingText +=
              string.Format(
                  "OFFER\tTID={0:X8}\tSIP={1}\t IP={2}\tLT={3}\tRT={4}\t{5}\t{6}\tDN={7}\tDNS={8}\tGW={9}",
                  xid, serverIp, yiaddr,
                  leaseTime, renewalTime, rebindingTime,
                  subnetMask,
                  domainName, dnsServers, routers);
          break;
        case 3:
          loggingText += string.Format("REQUEST\t{0:X8}\t{1}", xid, ciaddr);
          break;
        case 4:
          loggingText += string.Format("DECLINE\t{0:X8}\t{1}", xid, ciaddr);
          break;
        case 5:
          loggingText += string.Format("ACK\t{0:X8}\t{1}\t{2}", xid, ciaddr, yiaddr);
          break;
        case 6:
          loggingText += string.Format("NAK\t{0:X8}\t{1}\t{2}", xid, ciaddr, yiaddr);
          break;
        case 7:
          loggingText += string.Format("RELEASE\t{0:X8}\t{1}", xid, ciaddr);
          break;
        case 8:
          loggingText += string.Format("INFORM\t{0:X8}\t{1}\t{2}\t{3}", xid, ciaddr, yiaddr, hostName);
          break;
        default:
          loggingText += string.Format("Not handled: DHCP message type {0}", otype);
          break;
      }
    }

    public override string ToString()
    {
      return loggingText;
    }

    private static string ReadIpv4List(byte optionLength, BinaryReader binaryReader)
    {
      var sb = new StringBuilder();
      for (int i = 0; i < optionLength / 4; i++)
      {
        sb.Append(ReadIpv4(binaryReader));
        sb.Append(",");
      }
      sb.Remove(sb.Length - 1, 1);
      return sb.ToString();
    }

    private static string ReadIpv4(BinaryReader binaryReader)
    {
      byte[] ipv4 = binaryReader.ReadBytes(4);
      return String.Format("{0}.{1}.{2}.{3}", ipv4[0], ipv4[1], ipv4[2], ipv4[3]);
    }

    private static string ReadTimeInterval(BinaryReader binaryReader)
    {
      return Convert.ToString(
          (binaryReader.ReadByte() * Math.Pow(256, 3)) +
          (binaryReader.ReadByte() * Math.Pow(256, 2)) +
          (binaryReader.ReadByte() * 256) +
          binaryReader.ReadByte());
    }
  }
}
